from typing import Literal

import torch

from mipcandy.common import Normalize


def convert_ids_to_logits(ids: torch.Tensor, d: Literal[1, 2, 3], num_classes: int) -> torch.Tensor:
    if ids.dtype != torch.int or ids.min() < 0:
        raise TypeError("`ids` should be positive integers")
    d += 1
    if ids.ndim != d:
        if ids.ndim == d + 1 and ids.shape[1] == 1:
            ids = ids.squeeze(1)
        else:
            raise ValueError(f"`ids` should be {d} dimensional or {d + 1} dimensional with single channel")
    logits = torch.zeros((ids.shape[0], num_classes, *ids.shape[1:]), device=ids.device, dtype=torch.float32)
    logits.scatter_(1, ids.unsqueeze(1).long(), 1)
    return logits


def convert_logits_to_ids(logits: torch.Tensor, *, channel_dim: int = 1) -> torch.Tensor:
    return logits.round().int() if logits.shape[channel_dim] < 2 else logits.argmax(channel_dim, keepdim=True)


def auto_convert(image: torch.Tensor) -> torch.Tensor:
    return (image * 255 if 0 <= image.min() <= image.max() <= 1 else Normalize(domain=(0, 255))(image)).int()



from math import ceil
from typing import Literal

import torch
from torch import nn

from mipcandy.types import Colormap, Shape2d, Shape3d, Shape, Paddings2d, Paddings3d, Paddings


def reverse_paddings(paddings: Paddings) -> Paddings:
    if len(paddings) == 4:
        return paddings[2], paddings[3], paddings[0], paddings[1]
    return paddings[4], paddings[5], paddings[2], paddings[3], paddings[0], paddings[1]


class Pad(nn.Module):
    def __init__(self, *, value: int = 0, mode: str = "constant", batch: bool = True) -> None:
        super().__init__()
        self._value: int = value
        self._mode: str = mode
        self.batch: bool = batch
        self._paddings: Paddings | None = None
        self.requires_grad_(False)

    @staticmethod
    def _c_t(size: int, min_factor: int) -> int:
        """
        Compute target on a single dimension
        """
        return ceil(size / min_factor) * min_factor

    @staticmethod
    def _c_p(size: int, min_factor: int) -> tuple[int, int]:
        """
        Compute padding on a single dimension
        """
        excess = Pad._c_t(size, min_factor) - size
        before = excess // 2
        return before, excess - before


class Pad2d(Pad):
    def __init__(self, min_factor: int | Shape2d, *, value: int = 0, mode: str = "constant",
                 batch: bool = True) -> None:
        super().__init__(value=value, mode=mode, batch=batch)
        self._min_factor: Shape2d = (min_factor,) * 2 if isinstance(min_factor, int) else min_factor

    def paddings(self) -> Paddings2d | None:
        return self._paddings

    def padded_shape(self, in_shape: tuple[int, int, ...]) -> tuple[int, int, ...]:
        return *in_shape[:-2], self._c_t(in_shape[-2], self._min_factor[0]), self._c_t(
            in_shape[-1], self._min_factor[1])

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.batch:
            _, _, h, w = x.shape
            suffix = (0,) * 4
        else:
            _, h, w = x.shape
            suffix = (0,) * 2
        self._paddings = self._c_p(h, self._min_factor[0]) + self._c_p(w, self._min_factor[1])
        return nn.functional.pad(x, reverse_paddings(self._paddings) + suffix, self._mode, self._value)


class Pad3d(Pad):
    def __init__(self, min_factor: int | Shape3d, *, value: int = 0, mode: str = "constant",
                 batch: bool = True) -> None:
        super().__init__(value=value, mode=mode, batch=batch)
        self._min_factor: Shape3d = (min_factor,) * 3 if isinstance(min_factor, int) else min_factor

    def paddings(self) -> Paddings3d | None:
        return self._paddings

    def padded_shape(self, in_shape: tuple[int, int, int, ...]) -> tuple[int, int, int, ...]:
        return (*in_shape[:-3], self._c_t(in_shape[-3], self._min_factor[0]), self._c_t(
            in_shape[-2], self._min_factor[1]), self._c_t(in_shape[-1], self._min_factor[2]))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.batch:
            _, _, d, h, w = x.shape
            suffix = (0,) * 4
        else:
            _, d, h, w = x.shape
            suffix = (0,) * 2
        self._paddings = self._c_p(d, self._min_factor[0]) + self._c_p(h, self._min_factor[1]) + self._c_p(
            w, self._min_factor[2])
        return nn.functional.pad(x, reverse_paddings(self._paddings) + suffix, self._mode, self._value)


class Restore2d(nn.Module):
    def __init__(self, conjugate_padding: Pad2d) -> None:
        super().__init__()
        self.conjugate_padding: Pad2d = conjugate_padding
        self.requires_grad_(False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        paddings = self.conjugate_padding.paddings()
        if not paddings:
            raise ValueError("Paddings are not set yet, did you forget to pad before restoring?")
        pad_h0, pad_h1, pad_w0, pad_w1 = paddings
        if self.conjugate_padding.batch:
            _, _, h, w = x.shape
            return x[:, :, pad_h0: h - pad_h1, pad_w0: w - pad_w1]
        _, h, w = x.shape
        return x[:, pad_h0: h - pad_h1, pad_w0: w - pad_w1]


class Restore3d(nn.Module):
    def __init__(self, conjugate_padding: Pad3d) -> None:
        super().__init__()
        self.conjugate_padding: Pad3d = conjugate_padding
        self.requires_grad_(False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        paddings = self.conjugate_padding.paddings()
        if not paddings:
            raise ValueError("Paddings are not set yet, did you forget to pad before restoring?")
        pad_d0, pad_d1, pad_h0, pad_h1, pad_w0, pad_w1 = paddings
        if self.conjugate_padding.batch:
            _, _, d, h, w = x.shape
            return x[:, :, pad_d0: d - pad_d1, pad_h0: h - pad_h1, pad_w0: w - pad_w1]
        _, d, h, w = x.shape
        return x[:, pad_d0: d - pad_d1, pad_h0: h - pad_h1, pad_w0: w - pad_w1]


class PadTo(Pad):
    def __init__(self, min_shape: Shape, *, value: int = 0, mode: str = "constant", batch: bool = True) -> None:
        super().__init__(value=value, mode=mode, batch=batch)
        self._min_shape: Shape = min_shape
        self._pad: Pad2d | Pad3d = (Pad2d if len(min_shape) == 2 else Pad3d)(min_shape, value=value, mode=mode,
                                                                             batch=batch)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self._pad(x) if any(x.shape[i + (2 if self.batch else 1)] < min_size for i, min_size in enumerate(
            self._min_shape)) else x


class Normalize(nn.Module):
    def __init__(self, *, domain: tuple[float | None, float | None] = (0, None), strict: bool = False,
                 method: Literal["linear", "intercept", "cut", "zscore"] = "linear") -> None:
        super().__init__()
        self._domain: tuple[float | None, float | None] = domain
        self._strict: bool = strict
        self._method: Literal["linear", "intercept", "cut", "zscore"] = method
        self.requires_grad_(False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        left, right = self._domain
        if left is None and right is None and self._method != "zscore":
            return x
        r_l, r_r = x.min(), x.max()
        match self._method:
            case "linear":
                if left is None or (left < r_l and not self._strict):
                    left = r_l
                if right is None or (right > r_r and not self._strict):
                    right = r_r
                numerator = right - left
                if numerator == 0:
                    numerator = 1
                denominator = r_r - r_l
                if denominator == 0:
                    denominator = 1
                return (x - r_l) * numerator / denominator + left
            case "intercept":
                if left is not None and right is None:
                    return x - r_l + left if r_l < left or self._strict else x
                elif left is None and right is not None:
                    return x - r_r + right if r_r > right or self._strict else x
                else:
                    raise ValueError("Cannot use intercept normalization when both ends are fixed")
            case "cut":
                if self._strict:
                    raise ValueError("Method \"cut\" cannot be strict")
                if left is not None:
                    x = x.clamp(min=left)
                if right is not None:
                    x = x.clamp(max=right)
                return x
            case "zscore":
                if left is not None or right is not None:
                    raise ValueError("Method \"zscore\" cannot have fixed ends")
                return (x - x.mean()) / max(x.std(), torch.tensor(1e-8, device=x.device))


class CTNormalize(nn.Module):
    def __init__(self, mean_intensity: float, std_intensity: float, lower_bound: float, upper_bound: float) -> None:
        super().__init__()
        self._mean_intensity: float = mean_intensity
        self._std_intensity: float = std_intensity
        self._lower_bound: float = lower_bound
        self._upper_bound: float = upper_bound
        self.requires_grad_(False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return (x.clip(self._lower_bound, self._upper_bound) - self._mean_intensity) / max(self._std_intensity, 1e-8)


class ColorizeLabel(nn.Module):
    def __init__(self, *, colormap: Colormap | None = None, batch: bool = True) -> None:
        super().__init__()
        if not colormap:
            colormap = []
            for r in range(8):
                for g in range(8):
                    for b in range(32):
                        colormap.append([r * 32, g * 32, 255 - b * 32])
        self._colormap: torch.Tensor = torch.tensor(colormap)
        self._batch: bool = batch
        self.requires_grad_(False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self._batch:
            x = x.unsqueeze(0)
        cmap = self._colormap.to(x.device)
        x = (cmap[(x > 0).int()] if 0 <= x.min() < x.max() <= 1 else cmap[x.int()]).movedim(-1, 1)
        return x if self._batch else x.squeeze(0)



from typing import Literal

from mipcandy.config import load_settings, save_settings, load_secrets, save_secrets


def config(target: Literal["setting", "secret"], key: str, value: str) -> None:
    match target:
        case "setting":
            settings = load_settings()
            settings[key] = value
            save_settings(settings)
        case "secret":
            secrets = load_secrets()
            secrets[key] = value
            save_secrets(secrets)



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, asdict
from datetime import datetime
from hashlib import md5
from json import load, dump
from os import PathLike, urandom, makedirs, environ
from os.path import exists
from random import seed as random_seed, randint
from shutil import copy
from threading import Lock
from time import time
from typing import Sequence, override, Self

import numpy as np
import torch
from matplotlib import pyplot as plt
from pandas import DataFrame, read_csv
from rich.console import Console
from rich.progress import Progress, SpinnerColumn
from rich.table import Table
from torch import nn, optim
from torch.utils.data import DataLoader

from mipcandy.common import quotient_regression, quotient_derivative, quotient_bounds
from mipcandy.config import load_settings, load_secrets
from mipcandy.data import fast_save, fast_load, empty_cache
from mipcandy.frontend import Frontend
from mipcandy.layer import WithPaddingModule, WithNetwork
from mipcandy.profiler import Profiler
from mipcandy.sanity_check import sanity_check
from mipcandy.types import Params, Setting, AmbiguousShape


def try_append(new: float, to: dict[str, list[float]], key: str) -> None:
    if key in to:
        to[key].append(new)
    else:
        to[key] = [new]


def try_append_all(new: dict[str, float], to: dict[str, list[float]]) -> None:
    for key, value in new.items():
        try_append(value, to, key)


@dataclass
class TrainerToolbox(object):
    model: nn.Module
    optimizer: optim.Optimizer
    scheduler: optim.lr_scheduler.LRScheduler
    criterion: nn.Module
    ema: nn.Module | None = None


@dataclass
class TrainerTracker(object):
    epoch: int = 0
    best_score: float = float("-inf")
    worst_case: int | None = None


class Trainer(WithPaddingModule, WithNetwork, metaclass=ABCMeta):
    def __init__(self, trainer_folder: str | PathLike[str], dataloader: DataLoader[tuple[torch.Tensor, torch.Tensor]],
                 validation_dataloader: DataLoader[tuple[torch.Tensor, torch.Tensor]], *, recoverable: bool = True,
                 profiler: bool = False, device: torch.device | str = "cpu", console: Console = Console()) -> None:
        WithPaddingModule.__init__(self, device)
        WithNetwork.__init__(self, device)
        self._trainer_folder: str = trainer_folder
        self._trainer_variant: str = self.__class__.__name__
        self._experiment_id: str = "tbd"
        self._dataloader: DataLoader[tuple[torch.Tensor, torch.Tensor]] = dataloader
        self._validation_dataloader: DataLoader[tuple[torch.Tensor, torch.Tensor]] = validation_dataloader
        self._unrecoverable: bool | None = not recoverable  # None if the trainer is recovered
        self._console: Console = console
        self._metrics: dict[str, list[float]] = {}
        self._frontend: Frontend = Frontend({})
        self._lock: Lock = Lock()
        self._tracker: TrainerTracker = TrainerTracker()
        self._profiler: Profiler | None = None
        self._use_profiler: bool = profiler

    # Recovery methods (PR #108 at https://github.com/ProjectNeura/MIPCandy/pull/108)

    def save_everything_for_recovery(self, toolbox: TrainerToolbox, tracker: TrainerTracker,
                                     **training_arguments) -> None:
        if self._unrecoverable:
            return
        torch.save({
            "optimizer": toolbox.optimizer.state_dict(),
            "scheduler": toolbox.scheduler.state_dict(),
            "criterion": toolbox.criterion.state_dict()
        }, f"{self.experiment_folder()}/state_dicts.pth")
        with open(f"{self.experiment_folder()}/state_orb.json", "w") as f:
            dump({"tracker": asdict(tracker), "training_arguments": training_arguments}, f)

    def load_state_orb(self) -> dict[str, dict[str, Setting]]:
        with open(f"{self.experiment_folder()}/state_orb.json") as f:
            return load(f)

    def load_tracker(self) -> TrainerTracker:
        return TrainerTracker(**self.load_state_orb()["tracker"])

    def load_training_arguments(self) -> dict[str, Setting]:
        return self.load_state_orb()["training_arguments"]

    def load_metrics(self) -> dict[str, list[float]]:
        df = read_csv(f"{self.experiment_folder()}/metrics.csv", index_col="epoch")
        return {column: df[column].astype(float).tolist() for column in df.columns}

    def load_toolbox(self, num_epochs: int, example_shape: AmbiguousShape, compile_model: bool,
                     ema: bool) -> TrainerToolbox:
        toolbox = self._build_toolbox(num_epochs, example_shape, compile_model, ema, model=self.load_model(
            example_shape, compile_model, path=f"{self.experiment_folder()}/checkpoint_latest.pth"
        ))
        state_dicts = torch.load(f"{self.experiment_folder()}/state_dicts.pth")
        toolbox.optimizer.load_state_dict(state_dicts["optimizer"])
        toolbox.scheduler.load_state_dict(state_dicts["scheduler"])
        toolbox.criterion.load_state_dict(state_dicts["criterion"])
        return toolbox

    def recover_from(self, experiment_id: str) -> Self:
        self._experiment_id = experiment_id
        if not exists(self.experiment_folder()):
            raise FileNotFoundError(f"Experiment folder {self.experiment_folder()} not found")
        self._metrics = self.load_metrics()
        self._tracker = self.load_tracker()
        self._unrecoverable = None
        return self

    def continue_training(self, num_epochs: int) -> None:
        if not self.recovery():
            raise RuntimeError("Must call `recover_from()` before continuing training")
        self.train(num_epochs, **self.load_training_arguments())

    # Getters

    def trainer_folder(self) -> str:
        return self._trainer_folder

    def trainer_variant(self) -> str:
        return self._trainer_variant

    def experiment_id(self) -> str:
        return self._experiment_id

    def dataloader(self) -> DataLoader[tuple[torch.Tensor, torch.Tensor]]:
        return self._dataloader

    def validation_dataloader(self) -> DataLoader[tuple[torch.Tensor, torch.Tensor]]:
        return self._validation_dataloader

    def console(self) -> Console:
        return self._console

    def metrics(self) -> dict[str, list[float]]:
        return self._metrics.copy()

    def frontend(self) -> Frontend:
        return self._frontend

    def tracker(self) -> TrainerTracker:
        return self._tracker

    # Enhanced getters

    def initialized(self) -> bool:
        return self._experiment_id != "tbd"

    def recovery(self) -> bool:
        return self._unrecoverable is None

    def experiment_folder(self) -> str:
        return f"{self._trainer_folder}/{self._trainer_variant}/{self._experiment_id}"

    def get_example_input(self) -> torch.Tensor:
        return self._dataloader.dataset[0][0]

    def predict_maximum_validation_score(self, num_epochs: int, *, degree: int = 5) -> tuple[int, float]:
        val_scores = np.array(self._metrics["val score"])
        a, b = quotient_regression(np.arange(len(val_scores)), val_scores, degree, degree)
        da, db = quotient_derivative(a, b)
        max_roc = float(da[0] / db[0])
        max_val_score = float(a[0] / b[0])
        bounds = quotient_bounds(a, b, None, max_val_score * (1 - max_roc), x_start=0, x_stop=num_epochs, x_step=1)
        return (round(bounds[1]) + 1, max_val_score) if bounds else (0, 0)

    def etc(self, epoch: int, num_epochs: int, *, target_epoch: int | None = None,
            val_score_prediction_degree: int = 5) -> float:
        if not target_epoch:
            target_epoch, _ = self.predict_maximum_validation_score(num_epochs, degree=val_score_prediction_degree)
        epoch_durations = self._metrics["epoch duration"]
        return sum(epoch_durations) * (target_epoch - epoch) / len(epoch_durations)

    # Setters

    def set_frontend(self, frontend: type[Frontend], *, path_to_secrets: str | PathLike[str] | None = None) -> None:
        self._frontend = frontend(load_secrets(path=path_to_secrets) if path_to_secrets else load_secrets())

    def set_seed(self, seed: int) -> None:
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        random_seed(seed)
        np.random.seed(seed)
        environ['PYTHONHASHSEED'] = str(seed)
        if self.initialized():
            self.log(f"Set to manual seed {seed}")

    # Initialization methods

    def _allocate_experiment_folder(self) -> str:
        self._experiment_id = datetime.now().strftime("%Y%m%d-%H-") + md5(urandom(8)).hexdigest()[:4]
        experiment_folder = self.experiment_folder()
        return self._allocate_experiment_folder() if exists(experiment_folder) else experiment_folder

    def allocate_experiment_folder(self) -> str:
        return self.experiment_folder() if self.initialized() else self._allocate_experiment_folder()

    def init_experiment(self) -> None:
        if self.recovery():
            self.log(f"Training progress recovered from {self._experiment_id} from epoch {self._tracker.epoch}")
            return
        if self.initialized():
            raise RuntimeError("Experiment already initialized")
        makedirs(self._trainer_folder, exist_ok=True)
        experiment_folder = self.allocate_experiment_folder()
        makedirs(experiment_folder)
        t = datetime.now()
        with open(f"{experiment_folder}/logs.txt", "w") as f:
            f.write(f"File created by FightTumor, copyright (C) {t.year} Project Neura. All rights reserved\n")
        self.log(f"Experiment (ID {self._experiment_id}) created at {t}")
        self.log(f"Trainer: {self._trainer_variant}")
        if self._use_profiler:
            gpus = (self._device,) if torch.device(self._device).type == "cuda" else ()
            self._profiler = Profiler(self._trainer_variant, f"{experiment_folder}/profiler.txt", gpus=gpus)

    # Logging utilities

    def log(self, msg: str, *, on_screen: bool = True) -> None:
        msg = f"[{datetime.now()}] {msg}"
        if self.initialized():
            with open(f"{self.experiment_folder()}/logs.txt", "a") as f:
                f.write(f"{msg}\n")
        if on_screen:
            with self._lock:
                self._console.print(msg)

    def record(self, metric: str, value: float) -> None:
        try_append(value, self._metrics, metric)

    def record_all(self, metrics: dict[str, list[float]]) -> None:
        try_append_all({k: sum(v) / len(v) for k, v in metrics.items()}, self._metrics)

    def record_profiler(self) -> None:
        if self._profiler:
            self._profiler.record(stack_trace_offset=2)

    def record_profiler_linebreak(self, message: str) -> None:
        if self._profiler:
            self._profiler.line_break(message)
            self.log(f"[PROFILER] {message}")

    def save_metrics(self) -> None:
        df = DataFrame(self._metrics)
        df.index = range(1, len(df) + 1)
        df.index.name = "epoch"
        df.to_csv(f"{self.experiment_folder()}/metrics.csv")

    def save_metric_curve(self, name: str, values: Sequence[float]) -> None:
        name = name.capitalize()
        plt.plot(values)
        plt.title(f"{name} over Epoch")
        plt.xlabel("Epoch")
        plt.ylabel(name)
        plt.grid()
        plt.savefig(f"{self.experiment_folder()}/{name.lower()}.png")
        plt.close()

    def save_metric_curve_combo(self, metrics: dict[str, Sequence[float]], *, title: str = "All Metrics") -> None:
        for name, values in metrics.items():
            plt.plot(values, label=name.capitalize())
        plt.title(title)
        plt.xlabel("Epoch")
        plt.legend()
        plt.grid()
        plt.savefig(f"{self.experiment_folder()}/{title.lower()}.png")
        plt.close()

    def save_metric_curves(self, *, names: Sequence[str] | None = None) -> None:
        if names is None:
            for name, values in self._metrics.items():
                self.save_metric_curve(name, values)
        else:
            for name in names:
                self.save_metric_curve(name, self._metrics[name])

    def save_progress(self, *, names: Sequence[str] = ("combined loss", "val score")) -> None:
        self.save_metric_curve_combo({name: self._metrics[name] for name in names}, title="Progress")

    def save_preview(self, image: torch.Tensor, label: torch.Tensor, output: torch.Tensor, *,
                     quality: float = .75) -> None:
        ...

    def show_metrics(self, epoch: int, metrics: dict[str, list[float]], prefix: str, *, epochwise: bool = True,
                     lookup_prefix: str = "", global_previous_index: int = -2) -> None:
        """
        :param epoch: the current epoch number
        :param metrics: the metrics to show
        :param prefix: the prefix to use for the table title
        :param epochwise: whether the metrics are for one epoch (so that previous values are in `self._metrics`) or for all
            epochs (so that previous values are contained in :param: metrics itself)
        :param lookup_prefix: the prefix to use for the lookup in `self._metrics`
        :param global_previous_index: the index of the previous epoch in `self._metrics`
        """
        prefix = prefix.capitalize()
        table = Table(title=f"Epoch {epoch} {prefix}")
        table.add_column("Metric")
        table.add_column("Mean Value", style="green")
        table.add_column("Span", style="cyan")
        table.add_column("Diff", style="magenta")
        for metric, values in metrics.items():
            span = f"[{min(values):.4f}, {max(values):.4f}]"
            if epochwise:
                if global_previous_index >= 0:
                    raise ValueError("`global_previous_index` must be negative`")
                mean = sum(values) / len(values)
                value = f"{mean:.4f}"
                m = f"{lookup_prefix}{metric}"
                diff = f"{mean - self._metrics[m][global_previous_index]:+.4f}" if m in self._metrics and len(
                    self._metrics[m]) >= -global_previous_index else "N/A"
            else:
                value = f"{values[-1]:.4f}"
                diff = f"{values[-1] - values[-2]:+.4f}" if len(values) > 1 else "N/A"
            table.add_row(metric, value, span, diff)
            self.log(f"{prefix} {metric}: {value} @{span} ({diff})")
        self._console.print(table)

    def show_metrics_per_case(self, epoch: int, metrics: dict[str, list[float]]) -> None:
        table = Table(title=f"Epoch {epoch} Metrics per Case")
        table.add_column("Case ID")
        num_cases = 0
        for metric_name in metrics.keys():
            this_num_cases = len(metrics[metric_name])
            if not num_cases:
                num_cases = this_num_cases
            if this_num_cases != num_cases:
                raise ValueError(f"Expected {num_cases} cases for metric {metric_name}, got {this_num_cases}")
            table.add_column(metric_name, style="green")
        for i in range(num_cases):
            table.add_row(f"{i + 1}", *(f"{metrics[metric_name][i]:.4f}" for metric_name in metrics.keys()))
        self._console.print(table)

    # Builder interfaces

    @abstractmethod
    def build_optimizer(self, params: Params) -> optim.Optimizer:
        raise NotImplementedError

    @abstractmethod
    def build_scheduler(self, optimizer: optim.Optimizer, num_epochs: int) -> optim.lr_scheduler.LRScheduler:
        raise NotImplementedError

    @abstractmethod
    def build_criterion(self) -> nn.Module:
        raise NotImplementedError

    @abstractmethod
    def build_ema(self, model: nn.Module) -> nn.Module:
        raise NotImplementedError

    def _build_toolbox(self, num_epochs: int, example_shape: AmbiguousShape, compile_model: bool, ema: bool, *,
                       model: nn.Module | None = None) -> TrainerToolbox:
        if not model:
            model = self.load_model(example_shape, compile_model)
        optimizer = self.build_optimizer(model.parameters())
        scheduler = self.build_scheduler(optimizer, num_epochs)
        criterion = self.build_criterion().to(self._device)
        return TrainerToolbox(model, optimizer, scheduler, criterion, self.build_ema(model) if ema else None)

    def build_toolbox(self, num_epochs: int, example_shape: AmbiguousShape, compile_model: bool,
                      ema: bool) -> TrainerToolbox:
        return self._build_toolbox(num_epochs, example_shape, compile_model, ema)

    # Performance

    def empty_cache(self) -> None:
        empty_cache(self._device)

    # Training methods

    @abstractmethod
    def backward(self, images: torch.Tensor, labels: torch.Tensor, toolbox: TrainerToolbox) -> tuple[float, dict[
        str, float]]:
        raise NotImplementedError

    def train_batch(self, images: torch.Tensor, labels: torch.Tensor, toolbox: TrainerToolbox) -> tuple[float, dict[
        str, float]]:
        toolbox.optimizer.zero_grad()
        loss, metrics = self.backward(images, labels, toolbox)
        toolbox.optimizer.step()
        toolbox.scheduler.step()
        if toolbox.ema:
            toolbox.ema.update_parameters(toolbox.model)
        return loss, metrics

    def train_epoch(self, toolbox: TrainerToolbox) -> dict[str, list[float]]:
        self.record_profiler_linebreak(f"Epoch {self._tracker.epoch} training")
        self.record_profiler()
        self.record_profiler_linebreak("Emptying cache")
        self.empty_cache()
        self.record_profiler()
        toolbox.model.train()
        if toolbox.ema:
            toolbox.ema.train()
        metrics = {}
        with Progress(*Progress.get_default_columns(), SpinnerColumn(), console=self._console) as progress:
            task = progress.add_task(f"Epoch {self._tracker.epoch}", total=len(self._dataloader))
            for images, labels in self._dataloader:
                images, labels = images.to(self._device, non_blocking=True), labels.to(self._device, non_blocking=True)
                padding_module = self.get_padding_module()
                if padding_module:
                    images, labels = padding_module(images), padding_module(labels)
                progress.update(task, description=f"Training epoch {self._tracker.epoch} {tuple(images.shape)}")
                loss, batch_metrics = self.train_batch(images, labels, toolbox)
                try_append(loss, metrics, "combined loss")
                try_append_all(batch_metrics, metrics)
                progress.update(task, advance=1, description=f"Training epoch {self._tracker.epoch} ({loss:.4f})")
                self.record_profiler()
        return metrics

    def train(self, num_epochs: int, *, note: str = "", num_checkpoints: int = 5, compile_model: bool = True,
              ema: bool = True, seed: int | None = None, early_stop_tolerance: int = 5,
              val_score_prediction: bool = True, val_score_prediction_degree: int = 5, save_preview: bool = True,
              preview_quality: float = .75) -> None:
        training_arguments = self.filter_train_params(**locals())
        self.init_experiment()
        if note:
            self.log(f"Note: {note}")
        if seed is None:
            seed = randint(0, 100)
        self.set_seed(seed)
        self.record_profiler()
        self.record_profiler_linebreak("Sanity check")
        example_input = self.get_example_input().to(self._device).unsqueeze(0)
        padding_module = self.get_padding_module()
        if padding_module:
            example_input = padding_module(example_input)
        example_shape = tuple(example_input.shape[1:])
        self.log(f"Example input shape: {example_shape}")
        self.log("Building a template model to run sanity check on...")
        template_model = self.build_network(example_shape)
        model_name = template_model.__class__.__name__
        self.log(f"Model: {model_name}")
        sanity_check_result = sanity_check(template_model, example_shape, device=self._device)
        self.log(str(sanity_check_result))
        self.log(f"Example output shape: {tuple(sanity_check_result.output.shape)}")
        self.record_profiler()
        self.log("Building toolbox...")
        toolbox = (self.load_toolbox if self.recovery() else self.build_toolbox)(
            num_epochs, example_shape, compile_model, ema
        )
        checkpoint_path = lambda v: f"{self.experiment_folder()}/checkpoint_{v}.pth"
        es_tolerance = early_stop_tolerance
        self._frontend.on_experiment_created(self._experiment_id, self._trainer_variant, model_name, note,
                                             sanity_check_result.num_macs, sanity_check_result.num_params, num_epochs,
                                             early_stop_tolerance)
        del sanity_check_result, template_model, example_input
        self.empty_cache()
        try:
            for epoch in range(self._tracker.epoch, self._tracker.epoch + num_epochs):
                if early_stop_tolerance == -1:
                    epoch -= 1
                    self.log(f"Early stopping triggered because the validation score has not improved for {
                    es_tolerance} epochs")
                    break
                self._tracker.epoch = epoch
                # Training
                t0 = time()
                metrics = self.train_epoch(toolbox)
                self.record_all(metrics)
                lr = toolbox.scheduler.get_last_lr()[0]
                self.record("learning rate", lr)
                self.show_metrics(epoch, metrics, "training")
                self.save_model(toolbox.model, checkpoint_path("latest"))
                if epoch % (num_epochs / num_checkpoints) == 0:
                    copy(checkpoint_path("latest"), checkpoint_path(epoch))
                    self.log(f"Epoch {epoch} checkpoint saved")
                self.log(f"Epoch {epoch} training completed in {time() - t0:.1f} seconds")
                # Validation
                score, metrics = self.validate(toolbox)
                self.record_all({f"val {k}": v for k, v in metrics.items()})
                self.record("val score", score)
                msg = f"Validation score: {score:.4f}"
                if epoch > 1:
                    msg += f" ({score - self._metrics["val score"][-2]:+.4f})"
                self.log(msg)
                if val_score_prediction and epoch > val_score_prediction_degree:
                    target_epoch, max_score = self.predict_maximum_validation_score(
                        num_epochs, degree=val_score_prediction_degree
                    )
                    self.log(f"Maximum validation score {max_score:.4f} predicted at epoch {target_epoch}")
                    etc = self.etc(epoch, num_epochs, target_epoch=target_epoch)
                    self.log(f"Estimated time of completion in {etc:.1f} seconds at {datetime.fromtimestamp(
                        time() + etc):%m-%d %H:%M:%S}")
                self.show_metrics_per_case(epoch, metrics)
                self.show_metrics(epoch, metrics, "validation", lookup_prefix="val ")
                if score > self._tracker.best_score:
                    copy(checkpoint_path("latest"), checkpoint_path("best"))
                    self.log(f"======== Best checkpoint updated ({self._tracker.best_score:.4f} -> {
                    score:.4f}) ========")
                    self._tracker.best_score = score
                    early_stop_tolerance = es_tolerance
                    if save_preview:
                        self.save_preview(
                            fast_load(f"{self.experiment_folder()}/worst_input.pt"),
                            fast_load(f"{self.experiment_folder()}/worst_label.pt"),
                            fast_load(f"{self.experiment_folder()}/worst_output.pt"), quality=preview_quality
                        )
                else:
                    early_stop_tolerance -= 1
                epoch_duration = time() - t0
                self.record("epoch duration", epoch_duration)
                self.log(f"Epoch {epoch} completed in {epoch_duration:.1f} seconds")
                self.log(f"=============== Best Validation Score {self._tracker.best_score:.4f} ===============")
                self.save_metrics()
                self.save_progress()
                self.save_metric_curves()
                self.save_everything_for_recovery(toolbox, self._tracker, **training_arguments)
                self._frontend.on_experiment_updated(self._experiment_id, epoch, self._metrics, early_stop_tolerance)
        except Exception as e:
            self.log("Training interrupted")
            self.log(repr(e))
            self._frontend.on_experiment_interrupted(self._experiment_id, e)
            raise e
        else:
            self.log("Training completed")
            self._frontend.on_experiment_completed(self._experiment_id)

    @staticmethod
    def filter_train_params(**kwargs) -> dict[str, Setting]:
        return {k: v for k, v in kwargs.items() if k in (
            "note", "num_checkpoints", "compile_model", "ema", "seed", "early_stop_tolerance", "val_score_prediction",
            "val_score_prediction_degree", "save_preview", "preview_quality"
        )}

    def train_with_settings(self, num_epochs: int, **kwargs) -> None:
        settings = self.filter_train_params(**load_settings())
        settings.update(kwargs)
        self.train(num_epochs, **settings)

    # Validation methods

    @abstractmethod
    def validate_case(self, idx: int, image: torch.Tensor, label: torch.Tensor, toolbox: TrainerToolbox) -> tuple[
        float, dict[str, float], torch.Tensor]:
        raise NotImplementedError

    def validate(self, toolbox: TrainerToolbox) -> tuple[float, dict[str, list[float]]]:
        if self._validation_dataloader.batch_size != 1:
            raise RuntimeError("Validation dataloader should have batch size 1")
        self.record_profiler_linebreak(f"Validating epoch {self._tracker.epoch}")
        self.record_profiler()
        self.record_profiler_linebreak("Emptying cache")
        self.empty_cache()
        self.record_profiler()
        toolbox.model.eval()
        if toolbox.ema:
            toolbox.ema.eval()
        score = 0
        worst_score = float("+inf")
        metrics = {}
        num_cases = len(self._validation_dataloader)
        with torch.no_grad(), Progress(
                *Progress.get_default_columns(), SpinnerColumn(), console=self._console
        ) as progress:
            task = progress.add_task(f"Validating", total=num_cases)
            for idx, (image, label) in enumerate(self._validation_dataloader):
                image, label = image.to(self._device, non_blocking=True), label.to(self._device, non_blocking=True)
                padding_module = self.get_padding_module()
                if padding_module:
                    image, label = padding_module(image), padding_module(label)
                image, label = image.squeeze(0), label.squeeze(0)
                progress.update(task,
                                description=f"Validating epoch {self._tracker.epoch} case {idx} {tuple(image.shape)}")
                case_score, case_metrics, output = self.validate_case(idx, image, label, toolbox)
                score += case_score
                if case_score < worst_score:
                    self._tracker.worst_case = idx
                    fast_save(image, f"{self.experiment_folder()}/worst_input.pt")
                    fast_save(label, f"{self.experiment_folder()}/worst_label.pt")
                    fast_save(output, f"{self.experiment_folder()}/worst_output.pt")
                    worst_score = case_score
                try_append_all(case_metrics, metrics)
                progress.update(task, advance=1,
                                description=f"Validating epoch {self._tracker.epoch} case {idx} ({case_score:.4f})")
                self.record_profiler()
        return score / num_cases, metrics

    def __call__(self, *args, **kwargs) -> None:
        self.train(*args, **kwargs)

    @override
    def __str__(self) -> str:
        return f"{self._trainer_variant} {self._experiment_id}"



from dataclasses import dataclass, asdict
from json import dump, load
from math import ceil
from os import PathLike
from random import randint, choice
from typing import Sequence, override, Callable, Self, Any, Literal

import numpy as np
import torch
from rich.console import Console
from rich.progress import Progress, SpinnerColumn
from torch import nn

from mipcandy.data.dataset import SupervisedDataset
from mipcandy.data.geometric import crop
from mipcandy.types import Shape, AmbiguousShape


def format_bbox(bbox: Sequence[int]) -> tuple[int, int, int, int] | tuple[int, int, int, int, int, int]:
    if len(bbox) == 4:
        return bbox[0], bbox[1], bbox[2], bbox[3]
    elif len(bbox) == 6:
        return bbox[0], bbox[1], bbox[2], bbox[3], bbox[4], bbox[5]
    else:
        raise ValueError(f"Invalid bbox with {len(bbox)} elements")


@dataclass
class InspectionAnnotation(object):
    shape: AmbiguousShape
    foreground_bbox: tuple[int, int, int, int] | tuple[int, int, int, int, int, int]
    class_ids: tuple[int, ...]
    class_counts: dict[int, int]
    class_bboxes: dict[int, tuple[int, int, int, int] | tuple[int, int, int, int, int, int]]
    class_locations: dict[int, tuple[tuple[int, int] | tuple[int, int, int], ...]]
    spacing: Shape | None = None

    def foreground_shape(self) -> Shape:
        r = (self.foreground_bbox[1] - self.foreground_bbox[0], self.foreground_bbox[3] - self.foreground_bbox[2])
        return r if len(self.foreground_bbox) == 4 else r + (self.foreground_bbox[5] - self.foreground_bbox[4],)

    def center_of_foreground(self) -> tuple[int, int] | tuple[int, int, int]:
        r = (round((self.foreground_bbox[1] + self.foreground_bbox[0]) * .5),
             round((self.foreground_bbox[3] + self.foreground_bbox[2]) * .5))
        return r if len(self.shape) == 2 else r + (round((self.foreground_bbox[5] + self.foreground_bbox[4]) * .5),)


class InspectionAnnotations(Sequence[InspectionAnnotation]):
    def __init__(self, dataset: SupervisedDataset, background: int, intensity_stats: tuple[float, float, float, float],
                 *annotations: InspectionAnnotation) -> None:
        self._dataset: SupervisedDataset = dataset
        self._background: int = background
        self._intensity_stats: tuple[float, float, float, float] = intensity_stats
        self._annotations: tuple[InspectionAnnotation, ...] = annotations
        self._shapes: tuple[AmbiguousShape | None, AmbiguousShape, AmbiguousShape] | None = None
        self._foreground_shapes: tuple[AmbiguousShape | None, AmbiguousShape, AmbiguousShape] | None = None
        self._statistical_foreground_shape: Shape | None = None
        self._center_of_foregrounds: tuple[int, int] | tuple[int, int, int] | None = None
        self._foreground_offsets: tuple[int, int] | tuple[int, int, int] | None = None
        self._roi_shape: Shape | None = None

    def dataset(self) -> SupervisedDataset:
        return self._dataset

    def background(self) -> int:
        return self._background

    def intensity_stats(self) -> tuple[float, float, float, float]:
        """
        :return: mean, std, 0.5th percentile, 99.5th percentile
        """
        return self._intensity_stats

    def annotations(self) -> tuple[InspectionAnnotation, ...]:
        return self._annotations

    @override
    def __getitem__(self, item: int) -> InspectionAnnotation:
        return self._annotations[item]

    @override
    def __len__(self) -> int:
        return len(self._annotations)

    def save(self, path: str | PathLike[str]) -> None:
        with open(path, "w") as f:
            dump({
                "background": self._background, "intensity_stats": self._intensity_stats,
                "annotations": [asdict(a) for a in self._annotations]
            }, f)

    def _get_shapes(self, get_shape: Callable[[InspectionAnnotation], AmbiguousShape]) -> tuple[
        AmbiguousShape | None, AmbiguousShape, AmbiguousShape]:
        depths = []
        widths = []
        heights = []
        for annotation in self._annotations:
            shape = get_shape(annotation)
            if len(shape) == 2:
                heights.append(shape[0])
                widths.append(shape[1])
            else:
                depths.append(shape[0])
                heights.append(shape[1])
                widths.append(shape[2])
        return tuple(depths) if depths else None, tuple(heights), tuple(widths)

    def shapes(self) -> tuple[AmbiguousShape | None, AmbiguousShape, AmbiguousShape]:
        if self._shapes:
            return self._shapes
        self._shapes = self._get_shapes(lambda annotation: annotation.shape)
        return self._shapes

    def foreground_shapes(self) -> tuple[AmbiguousShape | None, AmbiguousShape, AmbiguousShape]:
        if self._foreground_shapes:
            return self._foreground_shapes
        self._foreground_shapes = self._get_shapes(lambda annotation: annotation.foreground_shape())
        return self._foreground_shapes

    def statistical_foreground_shape(self, *, percentile: float = .95) -> Shape:
        if self._statistical_foreground_shape:
            return self._statistical_foreground_shape
        depths, heights, widths = self.foreground_shapes()
        percentile *= 100
        sfs = (round(np.percentile(heights, percentile)), round(np.percentile(widths, percentile)))
        self._statistical_foreground_shape = (round(np.percentile(depths, percentile)),) + sfs if depths else sfs
        return self._statistical_foreground_shape

    def crop_foreground(self, i: int, *, expand_ratio: float = 1) -> tuple[torch.Tensor, torch.Tensor]:
        image, label = self._dataset.image(i), self._dataset.label(i)
        annotation = self._annotations[i]
        bbox = list(annotation.foreground_bbox)
        shape = annotation.foreground_shape()
        for dim_idx, size in enumerate(shape):
            left = int((expand_ratio - 1) * size // 2)
            right = int((expand_ratio - 1) * size - left)
            bbox[dim_idx * 2] = max(0, bbox[dim_idx * 2] - left)
            bbox[dim_idx * 2 + 1] = min(bbox[dim_idx * 2 + 1] + right, annotation.shape[dim_idx])
        return crop(image.unsqueeze(0), bbox).squeeze(0), crop(label.unsqueeze(0), bbox).squeeze(0)

    def foreground_heatmap(self) -> torch.Tensor:
        depths, heights, widths = self.foreground_shapes()
        max_shape = (max(depths), max(heights), max(widths)) if depths else (max(heights), max(widths))
        accumulated_label = torch.zeros((1, *max_shape), device=self._dataset.device())
        for i in range(len(self._dataset)):
            label = self._dataset.label(i)
            annotation = self._annotations[i]
            paddings = [0, 0, 0, 0]
            shape = annotation.foreground_shape()
            for j, size in enumerate(max_shape):
                left = (size - shape[j]) // 2
                right = size - shape[j] - left
                paddings.append(right)
                paddings.append(left)
            paddings.reverse()
            accumulated_label += nn.functional.pad(
                crop((label != self._background).unsqueeze(0), annotation.foreground_bbox), paddings
            ).squeeze(0)
        return accumulated_label.squeeze(0).detach()

    def center_of_foregrounds(self) -> tuple[int, int] | tuple[int, int, int]:
        if self._center_of_foregrounds:
            return self._center_of_foregrounds
        heatmap = self.foreground_heatmap()
        center = (heatmap.sum(dim=1).argmax().item(), heatmap.sum(dim=0).argmax().item()) if heatmap.ndim == 2 else (
            heatmap.sum(dim=(1, 2)).argmax().item(),
            heatmap.sum(dim=(0, 2)).argmax().item(),
            heatmap.sum(dim=(0, 1)).argmax().item(),
        )
        self._center_of_foregrounds = center
        return self._center_of_foregrounds

    def center_of_foregrounds_offsets(self) -> tuple[int, int] | tuple[int, int, int]:
        if self._foreground_offsets:
            return self._foreground_offsets
        center = self.center_of_foregrounds()
        depths, heights, widths = self.foreground_shapes()
        max_shape = (max(depths), max(heights), max(widths)) if depths else (max(heights), max(widths))
        offsets = (round(center[0] - max_shape[0] * .5), round(center[1] - max_shape[1] * .5))
        self._foreground_offsets = offsets + (round(center[2] - max_shape[2] * .5),) if depths else offsets
        return self._foreground_offsets

    def set_roi_shape(self, roi_shape: Shape | None) -> None:
        if roi_shape is not None:
            depths, heights, widths = self.shapes()
            if depths:
                if roi_shape[0] > min(depths) or roi_shape[1] > min(heights) or roi_shape[2] > min(widths):
                    raise ValueError(
                        f"ROI shape {roi_shape} exceeds minimum image shape ({min(depths)}, {min(heights)}, {min(widths)})")
            else:
                if roi_shape[0] > min(heights) or roi_shape[1] > min(widths):
                    raise ValueError(
                        f"ROI shape {roi_shape} exceeds minimum image shape ({min(heights)}, {min(widths)})")
        self._roi_shape = roi_shape

    def roi_shape(self, *, clamp: bool = True, percentile: float = .95) -> Shape:
        if self._roi_shape:
            return self._roi_shape
        sfs = self.statistical_foreground_shape(percentile=percentile)
        if clamp:
            if len(sfs) == 2:
                sfs = (None, *sfs)
            depths, heights, widths = self.shapes()
            roi_shape = (min(min(heights), sfs[1]), min(min(widths), sfs[2]))
            if depths:
                roi_shape = (min(min(depths), sfs[0]),) + roi_shape
            self._roi_shape = roi_shape
        else:
            self._roi_shape = sfs
        return self._roi_shape

    def roi(self, i: int, *, clamp: bool = True, percentile: float = .95) -> tuple[int, int, int, int] | tuple[
        int, int, int, int, int, int]:
        annotation = self._annotations[i]
        roi_shape = self.roi_shape(clamp=clamp, percentile=percentile)
        offsets = self.center_of_foregrounds_offsets()
        center = annotation.center_of_foreground()
        roi = []
        for i, position in enumerate(center):
            left = roi_shape[i] // 2
            right = roi_shape[i] - left
            offset = min(max(offsets[i], left - position), annotation.shape[i] - right - position)
            roi.append(position + offset - left)
            roi.append(position + offset + right)
        return tuple(roi)

    def crop_roi(self, i: int, *, clamp: bool = True, percentile: float = .95) -> tuple[torch.Tensor, torch.Tensor]:
        image, label = self._dataset.image(i), self._dataset.label(i)
        roi = self.roi(i, clamp=clamp, percentile=percentile)
        return crop(image.unsqueeze(0), roi).squeeze(0), crop(label.unsqueeze(0), roi).squeeze(0)


def _lists_to_tuples(pairs: Sequence[tuple[str, Any]]) -> dict[str, Any]:
    return {k: tuple(v) if isinstance(v, list) else v for k, v in pairs}


def _str_indices_to_int_indices(obj: dict[str, Any]) -> dict[int, Any]:
    return {int(k): v for k, v in obj.items()}


def parse_inspection_annotation(obj: dict[str, Any]) -> InspectionAnnotation:
    obj["class_bboxes"] = _str_indices_to_int_indices(obj["class_bboxes"])
    obj["class_locations"] = _str_indices_to_int_indices(obj["class_locations"])
    return InspectionAnnotation(**obj)


def load_inspection_annotations(path: str | PathLike[str], dataset: SupervisedDataset) -> InspectionAnnotations:
    with open(path) as f:
        obj = load(f, object_pairs_hook=_lists_to_tuples)
    annotations = InspectionAnnotations(dataset, obj["background"], obj["intensity_stats"], *(
        parse_inspection_annotation(row) for row in obj["annotations"]
    ))
    return annotations


def bbox_from_indices(indices: torch.Tensor, num_dim: Literal[2, 3]) -> tuple[int, int, int, int]:
    mins = indices.min(dim=0)[0].tolist()
    maxs = indices.max(dim=0)[0].tolist()
    bbox = (mins[1], maxs[1] + 1, mins[2], maxs[2] + 1)
    if num_dim == 3:
        bbox += (mins[3], maxs[3] + 1)
    return bbox


def inspect(dataset: SupervisedDataset, *, background: int = 0, max_samples: int = 10000,
            console: Console = Console()) -> InspectionAnnotations:
    r = []
    with torch.no_grad(), Progress(*Progress.get_default_columns(), SpinnerColumn(), console=console) as progress:
        task = progress.add_task("Inspecting dataset...", total=len(dataset))
        foreground_voxels = []
        for idx in range(len(dataset)):
            label = dataset.label(idx).int()
            progress.update(task, advance=1, description=f"Inspecting dataset {tuple(label.shape)}")
            ndim = label.ndim - 1
            fg_mask = label != background
            indices = fg_mask.nonzero()
            if len(indices) == 0:
                r.append(InspectionAnnotation(
                    tuple(label.shape[1:]), (0, 0, 0, 0) if ndim == 2 else (0, 0, 0, 0, 0, 0), (), {}, {}, {})
                )
                continue
            foreground_bbox = bbox_from_indices(indices, ndim)
            class_ids = label.unique().tolist()
            class_counts = {}
            class_bboxes = {}
            class_locations = {}
            for class_id in class_ids:
                indices = (label == class_id).nonzero()
                class_counts[class_id] = len(indices)
                class_bboxes[class_id] = bbox_from_indices(indices, ndim)
                if len(indices) > max_samples:
                    target_samples = min(max_samples, len(indices))
                    sampled_idx = torch.randperm(len(indices))[:target_samples]
                    indices = indices[sampled_idx]
                class_locations[class_id] = tuple(tuple(loc) for loc in indices[:, 1:].tolist())
            r.append(InspectionAnnotation(
                tuple(label.shape[1:]), foreground_bbox, tuple(
                    class_id for class_id in class_ids if class_id != background
                ), class_counts, class_bboxes, class_locations
            ))
            image = dataset.image(idx)
            if image.shape[0] > 1:
                fg_mask = fg_mask.expand_as(image)
            fg = image[fg_mask]
            if len(fg) > 0:
                foreground_voxels.append(fg)
        if len(foreground_voxels) == 0:
            raise ValueError("No foreground voxels found in dataset")
        all_fg = torch.cat(foreground_voxels)
        all_fg_np = all_fg.cpu().numpy()
        intensity_stats = (all_fg.mean().item(), all_fg.std().item(), float(np.percentile(all_fg_np, 0.5)),
                           float(np.percentile(all_fg_np, 99.5)))
    return InspectionAnnotations(dataset, background, intensity_stats, *r)


class ROIDataset(SupervisedDataset[list[int]]):
    def __init__(self, annotations: InspectionAnnotations, *, clamp: bool = True, percentile: float = .95) -> None:
        super().__init__(list(range(len(annotations))), list(range(len(annotations))),
                         transform=annotations.dataset().transform(), device=annotations.dataset().device())
        self._annotations: InspectionAnnotations = annotations
        self._clamp: bool = clamp
        self._percentile: float = percentile

    @override
    def construct_new(self, images: list[Any], labels: list[Any]) -> Self:
        new = self.__class__(self._annotations, percentile=self._percentile)
        new._images = images
        new._labels = labels
        return new

    @override
    def load_image(self, idx: int) -> torch.Tensor:
        raise NotImplementedError

    @override
    def load_label(self, idx: int) -> torch.Tensor:
        raise NotImplementedError

    @override
    def load(self, idx: int) -> tuple[torch.Tensor, torch.Tensor]:
        i = self._images[idx]
        if i != self._labels[idx]:
            raise ValueError(f"Image {i} and label {self._labels[idx]} indices do not match")
        with torch.no_grad():
            return self._annotations.crop_roi(i, clamp=self._clamp, percentile=self._percentile)


def crop_and_pad(x: torch.Tensor, bbox_lbs: list[int], bbox_ubs: list[int], *,
                 pad_value: int | float = 0) -> torch.Tensor:
    shape = x.shape[1:]
    dim = len(shape)
    valid_bbox_lbs = [max(0, bbox_lbs[i]) for i in range(dim)]
    valid_bbox_ubs = [min(shape[i], bbox_ubs[i]) for i in range(dim)]
    slices = tuple([slice(0, x.shape[0])] + [slice(valid_bbox_lbs[i], valid_bbox_ubs[i]) for i in range(dim)])
    cropped = x[slices]
    padding = [(-min(0, bbox_lbs[i]), max(bbox_ubs[i] - shape[i], 0)) for i in range(dim)]
    padding_torch = []
    for left, right in reversed(padding):
        padding_torch.extend([left, right])
    padded = nn.functional.pad(cropped, padding_torch, mode="constant", value=pad_value)
    return padded


class RandomROIDataset(ROIDataset):
    def __init__(self, annotations: InspectionAnnotations, batch_size: int, *, num_patches_per_case: int = 1,
                 oversample_rate: float = .33, clamp: bool = False, percentile: float = .5,
                 min_factor: int = 16) -> None:
        super().__init__(annotations, clamp=clamp, percentile=percentile)
        if num_patches_per_case > 1:
            images = [idx for idx in self._images for _ in range(num_patches_per_case)]
            self._images, self._labels = images, images.copy()
        self._batch_size: int = batch_size
        self._oversample_rate: float = oversample_rate
        sfs = self._annotations.statistical_foreground_shape(percentile=self._percentile)
        sfs = [ceil(s / min_factor) * min_factor for s in sfs]
        self._roi_shape: Shape = (min(sfs[0], 2048), min(sfs[1], 2048)) if len(sfs) == 2 else (
            min(sfs[0], 128), min(sfs[1], 128), min(sfs[2], 128))

    def convert_idx(self, idx: int) -> int:
        idx, idx2 = self._images[idx], self._labels[idx]
        if idx != idx2:
            raise ValueError(f"Image {idx} and label {idx2} indices do not match")
        return idx

    def roi_shape(self, *, roi_shape: Shape | None = None) -> None | Shape:
        if not roi_shape:
            return self._roi_shape
        self._roi_shape = roi_shape

    @override
    def construct_new(self, images: list[Any], labels: list[Any]) -> Self:
        new = self.__class__(self._annotations, self._batch_size, oversample_rate=self._oversample_rate,
                             clamp=self._clamp, percentile=self._percentile)
        new._images = images
        new._labels = labels
        new._roi_shape = self._roi_shape
        return new

    def random_roi(self, idx: int, force_foreground: bool) -> tuple[list[int], list[int]]:
        idx = self.convert_idx(idx)
        annotation = self._annotations[idx]
        roi_shape = self._roi_shape
        dim = len(annotation.shape)
        need_to_pad = [max(0, roi_shape[i] - annotation.shape[i]) for i in range(dim)]
        lbs = [-need_to_pad[i] // 2 for i in range(dim)]
        ubs = [annotation.shape[i] + need_to_pad[i] // 2 + need_to_pad[i] % 2 - roi_shape[i] for i in range(dim)]
        if force_foreground and len(annotation.class_ids) > 0:
            selected_class = choice(annotation.class_ids)
            selected_voxel = choice(annotation.class_locations[selected_class])
            bbox_lbs = [max(lbs[i], selected_voxel[i] - roi_shape[i] // 2) for i in range(dim)]
        else:
            bbox_lbs = [randint(lbs[i], ubs[i]) for i in range(dim)]
        return bbox_lbs, [bbox_lbs[i] + roi_shape[i] for i in range(dim)]

    def oversample_foreground(self, idx: int) -> bool:
        return idx % self._batch_size >= round(self._batch_size * (1 - self._oversample_rate))

    @override
    def load(self, idx: int) -> tuple[torch.Tensor, torch.Tensor]:
        force_foreground = self.oversample_foreground(idx)
        lbs, ubs = self.random_roi(idx, force_foreground)
        dataset = self._annotations.dataset()
        idx = self.convert_idx(idx)
        return crop_and_pad(dataset.image(idx), lbs, ubs), crop_and_pad(dataset.label(idx), lbs, ubs)



from dataclasses import dataclass
from inspect import stack
from os import PathLike
from time import time
from typing import Sequence, override

import torch
from psutil import cpu_percent, virtual_memory

from mipcandy.types import Device


@dataclass
class ProfilerFrame(object):
    stack: str
    cpu: float
    mem: float
    gpu: list[float] | None = None
    gpu_mem: list[float] | None = None

    @override
    def __str__(self) -> str:
        r = f"[{self.stack}] CPU: {self.cpu:.2f}% @ Memory: {self.mem:.2f}%\n"
        if self.gpu and self.gpu_mem:
            for i, gpu in enumerate(self.gpu):
                r += f"\t\tGPU {i}: {gpu:.2f}% @ Memory: {self.gpu_mem[i]:.2f}%\n"
        return r

    def export(self, duration: float) -> str:
        return f"{duration:.2f}s\t{self}"


class _LineBreak(object):
    def __init__(self, message: str) -> None:
        self.message: str = message

    @override
    def __str__(self) -> str:
        return f"<{self.message}>\n"

    def export(self, duration: float) -> str:
        return f"{duration:.2f}s\t{self}"


class Profiler(object):
    def __init__(self, title: str, save_as: str | PathLike[str], *, gpus: Sequence[Device] = ()) -> None:
        self.title: str = title
        self.save_as: str = save_as
        self.total_mem: float = self.get_total_mem()
        self.has_gpu: bool = len(gpus) > 0
        self._gpus: Sequence[Device] = gpus
        self.total_gpu_mem: list[float] = [self.get_total_gpu_mem(device) for device in gpus]
        with open(save_as, "w") as f:
            f.write(f"# {title}\nTotal memory: {self.total_mem}, Total GPU memory: {self.total_gpu_mem}\n\n")
        self._t0: float = time()

    @staticmethod
    def get_cpu_usage() -> float:
        return cpu_percent()

    def get_mem_usage(self) -> float:
        return 100 * virtual_memory().used / self.total_mem

    @staticmethod
    def get_total_mem() -> float:
        return virtual_memory().total

    @staticmethod
    def get_gpu_usage(device: Device) -> float:
        return torch.cuda.utilization(device)

    def get_gpu_mem_usage(self, device: Device) -> float:
        return 100 * torch.cuda.device_memory_used(device) / self.total_gpu_mem[self._gpus.index(device)]

    @staticmethod
    def get_total_gpu_mem(device: Device) -> float:
        return torch.cuda.get_device_properties(device).total_memory

    def _save(self, obj: ProfilerFrame | _LineBreak) -> None:
        with open(self.save_as, "a") as f:
            t = time()
            f.write(f"{obj.export(t - self._t0)}\n")
            self._t0 = t

    def record(self, *, stack_trace_offset: int = 1) -> ProfilerFrame:
        frame = ProfilerFrame(" -> ".join([f"{f.function}:{f.lineno}" for f in reversed(stack()[stack_trace_offset:])]),
                              self.get_cpu_usage(), self.get_mem_usage())
        if self.has_gpu:
            frame.gpu = [torch.cuda.utilization(device) for device in self._gpus]
            frame.gpu_mem = [self.get_gpu_mem_usage(device) for device in self._gpus]
        self._save(frame)
        return frame

    def line_break(self, message: str) -> None:
        self._save(_LineBreak(message))



import torch
from torch import nn

from mipcandy.types import Transform


class JointTransform(nn.Module):
    def __init__(self, *, transform: Transform | None = None, image_only: Transform | None = None,
                 label_only: Transform | None = None, keys: tuple[str, str] = ("image", "label")) -> None:
        super().__init__()
        self.transform: Transform | None = transform
        self.image_only: Transform | None = image_only
        self.label_only: Transform | None = label_only
        self._keys: tuple[str, str] = keys

    def forward(self, image: torch.Tensor, label: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        ik, lk = self._keys
        data = {ik: image, lk: label}
        if self.transform:
            data = self.transform(data)
        if self.image_only:
            data[ik] = self.image_only(data[ik])
        if self.label_only:
            data[lk] = self.label_only(data[lk])
        return data[ik], data[lk]


class MONAITransform(nn.Module):
    def __init__(self, transform: Transform, *, keys: tuple[str, str] = ("image", "label")) -> None:
        super().__init__()
        self.transform: Transform = transform
        self._keys: tuple[str, str] = keys

    def forward(self, data: torch.Tensor | dict[str, torch.Tensor]) -> torch.Tensor | dict[str, torch.Tensor]:
        if isinstance(data, torch.Tensor):
            return self.transform(data)
        ik, lk = self._keys
        image, label = data[ik], data[lk]
        return {ik: self.transform(image), lk: self.transform(label)}



from os import PathLike
from zipfile import ZipFile

from requests import get
from rich.console import Console
from rich.progress import track


def download_dataset(name: str, to: str | PathLike[str], *, endpoint: str = "cds.projectneura.org",
                     console: Console = Console()) -> None:
    to_zip = f"{to}.zip"
    with get(f"https://{endpoint}/{name}.zip", stream=True) as response:
        response.raise_for_status()
        with open(to_zip, "wb") as f:
            for chunk in track(response.iter_content(chunk_size=8192), description="Downloading...", console=console):
                f.write(chunk)
    console.print(f"Dataset successfully downloaded as {to_zip}")
    console.print("Unzipping...")
    with ZipFile(to_zip, "r") as zip_ref:
        zip_ref.extractall(to)
    console.print(f"Dataset successfully extracted to {to}")



from importlib.util import find_spec
from os import PathLike
from typing import Any, Iterable, Sequence, Literal

import torch
from torch import nn
from torchvision.transforms import Compose

type Setting = str | int | float | bool | None | dict[str, Setting] | list[Setting]
type Settings = dict[str, Setting]
type Params = Iterable[torch.Tensor] | Iterable[dict[str, Any]]
if find_spec("monai"):
    from monai.transforms import Transform as _Transform

    type Transform = nn.Module | Compose | _Transform
else:
    type Transform = nn.Module | Compose
type SupportedPredictant = Sequence[torch.Tensor] | str | PathLike[str] | Sequence[str] | torch.Tensor
type Colormap = Sequence[int | tuple[int, int, int]]
type Device = torch.device | str
type Shape2d = tuple[int, int]
type Shape3d = tuple[int, int, int]
type Shape = Shape2d | Shape3d
type AmbiguousShape = tuple[int, ...]
type Paddings2d = tuple[int, int, int, int]
type Paddings3d = tuple[int, int, int, int, int, int]
type Paddings = Paddings2d | Paddings3d
type Reduction = Literal["mean", "median", "sum", "none"]



from gc import collect
from math import floor
from os import PathLike

import SimpleITK as SpITK
import torch
from safetensors.torch import save_file, load_file

from mipcandy.data.convertion import auto_convert
from mipcandy.data.geometric import ensure_num_dimensions
from mipcandy.types import Device


def fast_save(x: torch.Tensor, path: str | PathLike[str]) -> None:
    save_file({"payload": x if x.is_contiguous() else x.contiguous()}, path)


def fast_load(path: str | PathLike[str], *, device: Device = "cpu") -> torch.Tensor:
    return load_file(path, device)["payload"]


def resample_to_isotropic(image: SpITK.Image, *, target_iso: float | None = None,
                          interpolator: int = SpITK.sitkBSpline) -> SpITK.Image:
    dim = image.GetDimension()
    old_spacing = image.GetSpacing()
    old_size = image.GetSize()
    origin = image.GetOrigin()
    direction = image.GetDirection()
    if target_iso is None:
        target_iso = min(old_spacing)
    new_spacing = (target_iso,) * dim
    new_size = tuple(max(1, floor(old_spacing[i] * (old_size[i] - 1) / new_spacing[i] + 1)) for i in range(dim))
    return SpITK.Resample(
        image, new_size, SpITK.Transform(), interpolator, origin, new_spacing, direction, 0, image.GetPixelID()
    )


def load_image(path: str | PathLike[str], *, is_label: bool = False, align_spacing: bool = False,
               target_iso: float | None = None, device: Device = "cpu") -> torch.Tensor:
    file = SpITK.ReadImage(path)
    if align_spacing:
        file = resample_to_isotropic(file, target_iso=target_iso,
                                     interpolator=SpITK.sitkNearestNeighbor if is_label else SpITK.sitkBSpline)
    img = torch.tensor(SpITK.GetArrayFromImage(file), dtype=torch.long if is_label else torch.float, device=device)
    if path.endswith(".nii.gz") or path.endswith(".nii") or path.endswith(".mha"):
        img = ensure_num_dimensions(img, 4, append_before=False).permute(3, 0, 1, 2)
        return img.squeeze(1) if img.shape[1] == 1 else img
    if path.endswith(".png") or path.endswith(".jpg") or path.endswith(".jpeg"):
        return ensure_num_dimensions(img, 3, append_before=False).permute(2, 0, 1)
    raise NotImplementedError(f"Unsupported file type: {path}")


def save_image(image: torch.Tensor, path: str | PathLike[str]) -> None:
    if path.endswith(".nii.gz") or path.endswith(".nii") or path.endswith(".mha"):
        image = ensure_num_dimensions(image, 4).permute(1, 2, 3, 0)
        return SpITK.WriteImage(SpITK.GetImageFromArray(image.detach().cpu().numpy(), isVector=False), path)
    if path.endswith(".png") or path.endswith(".jpg") or path.endswith(".jpeg"):
        image = auto_convert(ensure_num_dimensions(image, 3)).to(torch.uint8).permute(1, 2, 0)
        return SpITK.WriteImage(SpITK.GetImageFromArray(image.detach().cpu().numpy(), isVector=True), path)
    raise NotImplementedError(f"Unsupported file type: {path}")


def empty_cache(device: Device) -> None:
    match torch.device(device).type:
        case "cpu":
            collect()
        case "cuda":
            torch.cuda.empty_cache()
        case "mps":
            torch.mps.empty_cache()
